# Make marshmallow's functions and classes importable from flask-io
from marshmallow import pre_load, pre_dump, post_load, post_dump, Schema, ValidationError
from marshmallow.utils import missing

from .io import FlaskIO
from .utils import Error

__version__ = '1.7.0'
__author__ = 'Vinicius Chiele'
__license__ = 'MIT'



"""
Custom Field classes that extend marshmallow Field class.

All fields from marshmallow has been imported here to allow the user import them from the flask-io.
"""

from marshmallow.fields import *
from marshmallow.validate import Length, OneOf
from .validate import Complexity


class DelimitedList(List):
    """
    A delimited list composed with another `Field` class that loads from a delimited string.
    """

    delimiter = ','

    def __init__(self, cls_or_instance, delimiter=None, **kwargs):
        """
        Initializes a new instance of `DelimitedList`.

        :param Field cls_or_instance: A field class or instance.
        :param str delimiter: Delimiter between values.
        """
        self.delimiter = delimiter or self.delimiter
        super().__init__(cls_or_instance, **kwargs)

    def _serialize(self, value, attr, obj):
        values = super()._serialize(value, attr, obj)
        return self.delimiter.join(format(v) for v in values)

    def _deserialize(self, value, attr, data):
        values = value.split(self.delimiter)
        return super()._deserialize(values, attr, data)


class Enum(Field):
    """
    A field that provides a set of enumerated values which an attribute must be constrained to.
    """

    def __init__(self, enum_type, *args, **kwargs):
        """
        Initializes a new instance of `Enum`.

        :param enum.Enum enum_type: A Python enum class.
        """

        super().__init__(*args, **kwargs)
        self.enum_type = enum_type
        self.__member_type = type(list(self.enum_type)[0].value)

        self.validators.append(OneOf([v.value for v in self.enum_type]))

    def _serialize(self, value, attr, obj):
        if type(value) is self.enum_type:
            return value.value
        if type(value) is not self.__member_type:
            value = self.__member_type(value)
        return self.enum_type(value).value

    def _deserialize(self, value, attr, data):
        if type(value) is self.enum_type:
            return value
        if type(value) is not self.__member_type:
            value = self.__member_type(value)
        return self.enum_type(value)

    def _validate(self, value):
        if type(value) is self.enum_type:
            super()._validate(value.value)
        else:
            super()._validate(value)


class Password(Field):
    """
    A password field used to validate strong passwords.
    """
    def __init__(self, upper=1, lower=1, letters=1, digits=1, special=1, special_chars=None, min_length=6, max_length=None, *args, **kwargs):
        """
        Initializes a new instance of `Password`.

        :param int upper: Number of uppercase letters that the password must have.
        :param int lower: Number of lowercase letters that the password must have.
        :param int letters: Number of letters that the password must have.
        :param int letters: Number of special characters that the password must have.
        :param str special_chars: List of special characters allowed.
        :param int min_length: Minimum of characters that the password must have.
        :param int max_length: Maximum of characters that the password must have.
        """
        super().__init__(*args, **kwargs)

        self.validators.append(Length(min_length, max_length))
        self.validators.append(Complexity(upper, lower, letters, digits, special, special_chars))



from flask import request
from time import perf_counter
from werkzeug.http import HTTP_STATUS_CODES


def errors_to_dict(errors):
    if isinstance(errors, str):
        errors = [Error(errors)]
    elif not isinstance(errors, list):
        errors = [errors]

    errors_data = []

    for error in errors:
        errors_data.append(error.as_dict())

    return dict(errors=errors_data)


def format_trace_data(data):
    request_method = data.pop('request_method', None)
    request_url = data.pop('request_url', None)
    latency = data.pop('latency', None)
    request_headers = data.pop('request_headers', None)
    request_body = data.pop('request_body', None)
    response_status = data.pop('response_status', None)
    error = data.pop('error', None)

    message = ''

    if request_method:
        message += request_method + ' '

    if request_url:
        message += request_url + ' '

    if response_status:
        message += str(response_status) + ' '

    if latency:
        message += '%.5f' % latency

    message += '\r\n'

    for key, value in data.items():
        message += key + ': ' + str(value) + '\r\n'

    if request_headers:
        for key, value in request_headers.items():
            message += key + ': ' + str(value) + '\r\n'

    if error:
        message += '\r\n' + error
    elif request_body:
        message += '\r\n' + request_body

    return message


def get_best_match_for_content_type(mimetypes, default=None):
    content_type = request.headers['content-type']

    if not content_type:
        return default

    mimetype_expected = content_type.split(';')[0].lower()
    for mimetype in mimetypes:
        if mimetype_expected == mimetype:
            return mimetype
    return None


def http_status_message(code):
    return HTTP_STATUS_CODES.get(code, '')


def marshal(data, schema, envelope=None):
    if data is not None:
        many = isinstance(data, list)
        data = schema.dump(data, many=many).data

    if envelope:
        return {envelope: data}

    return data


def unpack(value):
    data, status, headers = value + (None,) * (3 - len(value))
    return data, status, headers


def validation_error_to_errors(validation_error):
    errors = []

    for field, error in validation_error.messages.items():
        validation_error_to_error(field, error, validation_error.kwargs.get('location'), errors)

    return errors


def validation_error_to_error(field, error, location, errors):
    if isinstance(error, dict):
        for f, e in error.items():
            validation_error_to_error(f, e, location, errors)
    elif isinstance(error, list):
        error = error[0]
        if isinstance(error, str):
            errors.append(Error(error, location=location, field=field))
        elif isinstance(error, dict):
            errors.append(Error(error.get('message'), error.get('code'), location, field))


class Error(object):
    def __init__(self, message, code=None, location=None, field=None, **kwargs):
        self.message = message
        self.code = code
        self.location = location
        self.field = field
        self.__dict__.update(kwargs)

    def as_dict(self):
        data = self.__dict__.copy()

        for key in set(data.keys()):
            if data[key] is None:
                data.pop(key)

        return data


class Stopwatch(object):
    def __init__(self):
        self.elapsed = 0.0
        self._start = None

    @staticmethod
    def start_new():
        sw = Stopwatch()
        sw.start()
        return sw

    def start(self):
        if not self._start:
            self._start = perf_counter()

    def stop(self):
        if self._start:
            end = perf_counter()
            self.elapsed += (end - self._start)
            self._start = None

    def reset(self):
        self.elapsed = 0.0

    @property
    def running(self):
        return self._start is not None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()



"""
Tracing classes.
"""

from collections import OrderedDict
from .utils import format_trace_data


class Tracer(object):
    """
    Responsible to collect and emit the tracing samples.
    """

    def __init__(self, io):
        self.io = io
        self.filters = []
        self.enabled = False
        self.inspector = lambda data: None
        self.emitter = self.__default_emit_trace

    def add_filter(self, methods=None, endpoints=None):
        """
        Adds a filter.

        :param methods: The http methods to be filtered.
        :param endpoints: The endpoints to be filtered.
        :return:
        """
        if not methods and not endpoints:
            raise ValueError('Filter cannot be added with no criteria.')

        filter = TraceFilter(methods, endpoints)
        self.filters.append(filter)
        return filter

    def match(self, rule):
        """
        Checks if the specified rule matches with any filter added.
        :param rule: The Flask rule to be matched.
        :return: True if there is a filter that matches.
        """
        if len(self.filters) == 0:
            return True

        for filter in self.filters:
            if filter.match(rule):
                return True
        return False

    def trace(self, request, response, error, latency):
        """
        Collects the data from the specified parameters and emit it.

        :param request: The Flask request object.
        :param response: The Flask response object.
        :param error: The error occurred if any.
        :param latency: The time elapsed to process the request.
        """

        data = self.__collect_trace_data(request, response, error, latency)

        self.inspector(data)
        self.emitter(data)

    def __collect_trace_data(self, request, response, error, latency):
        data = OrderedDict()
        data['latency'] = latency.elapsed
        data['request_method'] = request.environ['REQUEST_METHOD']
        data['request_url'] = request.url
        data['request_headers'] = request.headers

        body = request.get_data(as_text=True)
        if body:
            data['request_body'] = body

        if response:
            data['response_status'] = response.status_code

        if error:
            data['error'] = str(error)

        return data

    def __default_emit_trace(self, data):
        message = format_trace_data(data)
        self.io.logger.info(message)


class TraceFilter(object):
    """
    A tracing filter.
    """

    def __init__(self, methods, endpoints):
        self.methods = methods
        self.endpoints = endpoints

    def match(self, rule):
        """
        Checks if the specified rule matches with the filter.
        :param rule: The Flask rule to be matched.
        :return: True if the filter matches.
        """
        if self.methods:
            for method in self.methods:
                if method in rule.methods:
                    return True

        if self.endpoints:
            for endpoint in self.endpoints:
                if endpoint == rule.endpoint:
                    return True

        return False



"""
Custom Validator classes that extend marshmallow Validator class.

All validators from marshmallow has been imported here to allow the user import them from the flask-io.
"""

import string

from marshmallow.validate import *


class Complexity(Validator):
    """
    A validator that allows to validate a str in several ways.

    The code below was inspired by django-password library.
    https://github.com/dstufft/django-passwords/blob/master/passwords/validators.py
    """

    # this has been added to prevent an issue with unit test.
    error = None

    def __init__(self, upper=0, lower=0, letters=0, digits=0, special=0, special_chars=None):
        """
        Initializes a new instance of `Complexity`.

        :param int upper: Number of uppercase letters that the str must have.
        :param int lower: Number of lowercase letters that the str must have.
        :param int letters: Number of letters that the str must have.
        :param int letters: Number of special characters that the str must have.
        :param str special_chars: List of special characters allowed.
        """
        self.upper = upper
        self.lower = lower
        self.letters = letters
        self.digits = digits
        self.special = special
        self.special_chars = special_chars or string.punctuation

    def __call__(self, value):
        uppercase, lowercase, letters = set(), set(), set()
        digits, special = set(), set()

        for character in value:
            if character.isupper():
                uppercase.add(character)
                letters.add(character)
            elif character.islower():
                lowercase.add(character)
                letters.add(character)
            elif character.isdigit():
                digits.add(character)
            elif character in self.special_chars:
                special.add(character)
            else:
                raise ValidationError('Only %s are allowed as special characters' % self.special_chars)

        if len(uppercase) < self.upper:
            raise ValidationError('Must contain %s or more unique uppercase characters' % self.upper)

        if len(lowercase) < self.lower:
            raise ValidationError('Must contain %s or more unique lowercase characters' % self.lower)

        if len(letters) < self.letters:
            raise ValidationError('Must contain %s or more unique lowercase letters' % self.letters)

        if len(digits) < self.digits:
            raise ValidationError('Must contain %s or more unique digits' % self.digits)

        if len(special) < self.special:
            raise ValidationError('Must contain %s or more unique special characters' % self.special)

        return value



import functools

from collections import OrderedDict
from flask import request
from inspect import isclass
from logging import getLogger
from werkzeug.exceptions import BadRequest, InternalServerError, HTTPException, UnsupportedMediaType
from . import fields, missing, ValidationError
from .encoders import json_decode, json_encode
from .tracing import Tracer
from .utils import get_best_match_for_content_type, errors_to_dict, http_status_message
from .utils import marshal, unpack, validation_error_to_errors, Stopwatch


class FlaskIO(object):
    """
    """

    # default mime type for decode
    default_decoder = 'application/json'

    # default mime type for encode
    default_encoder = 'application/json'

    def __init__(self, app=None):
        self.__app = None

        self.decoders = OrderedDict([('application/json', json_decode)])
        self.encoders = OrderedDict([('application/json', json_encode)])

        self.logger = getLogger('flask-io')

        self.tracer = Tracer(self)

        if app:
            self.init_app(app)

    def init_app(self, app):
        """
        Initialize this class with the specified :class:`flask.Flask` application

        :param app: The Flask application.
        """

        self.__app = app
        self.__app.before_first_request(self.__setup)

        self.tracer.enabled = self.__app.config.get('TRACE_ENABLED', self.tracer.enabled)

    def bad_request(self, error):
        """
        Gets a 404 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.make_response((errors_to_dict(error), 400))

    def conflict(self, error):
        """
        Gets a 409 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.make_response((errors_to_dict(error), 409))

    def created(self, data, schema=None, envelope=None):
        """
        Gets a 201 response with the specified data.

        :param data: The content value.
        :param schema: The schema to serialize the data.
        :param data: The key used to envelope the data.
        :return: A Flask response object.
        """

        data = marshal(data, schema, envelope)
        return self.make_response((data, 201))

    def forbidden(self, error):
        """
        Gets a 403 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.make_response((errors_to_dict(error), 403))

    def no_content(self):
        """
        Gets a 204 response with no content.

        :return: A Flask response object.
        """

        return self.make_response((None, 204))

    def not_found(self, error):
        """
        Gets a 404 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """

        return self.make_response((errors_to_dict(error), 404))

    def ok(self, data, schema=None, envelope=None):
        """
        Gets a 200 response with the specified data.

        :param data: The content value.
        :param schema: The schema to serialize the data.
        :param data: The key used to envelope the data.
        :return: A Flask response object.
        """

        data = marshal(data, schema, envelope)
        return self.make_response(data)

    def unauthorized(self, error):
        """
        Gets a 401 response with the specified error.

        :param error: The error to include in the response.
        :return: A Flask response object.
        """
        return self.make_response((errors_to_dict(error), 401))

    def decoder(self, media_type):
        """
        A decorator that sets a decoder for the specified media type.

        :param media_type: The media type
        :return: A function
        """
        def wrapper(func):
            self.decoders[media_type] = func
            return func
        return wrapper

    def encoder(self, media_type):
        """
        A decorator that sets a encoder for the specified media type.

        :param media_type: The media type
        :return: A function
        """
        def wrapper(func):
            self.encoders[media_type] = func
            return func
        return wrapper

    def from_body(self, param_name, schema):
        """
        A decorator that converts the request body into a function parameter based on the specified schema.

        :param param_name: The parameter which receives the argument.
        :return: A function
        """

        schema = schema() if isclass(schema) else schema

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                kwargs[param_name] = self.__parse_body(schema)
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def from_cookie(self, param_name, field):
        """
        A decorator that converts a request cookie into a function parameter based on the specified field.

        :param param_name: The parameter which receives the argument.
        :return: A function
        """

        return self.__from_source(param_name, field, lambda: request.cookies, 'cookie')

    def from_form(self, param_name, field):
        """
        A decorator that converts a request form into a function parameter based on the specified field.

        :param param_name: The parameter which receives the argument.
        :return: A function
        """
        return self.__from_source(param_name, field, lambda: request.form, 'form')

    def from_header(self, param_name, field):
        """
        A decorator that converts a request header into a function parameter based on the specified field.

        :param param_name: The parameter which receives the argument.
        :return: A function
        """
        return self.__from_source(param_name, field, lambda: request.headers, 'header')

    def from_query(self, param_name, field):
        """
        A decorator that converts a query string into a function parameter based on the specified field.

        :param param_name: The parameter which receives the argument.
        :return: A function
        """

        return self.__from_source(param_name, field, lambda: request.args, 'query')

    def marshal_with(self, schema, envelope=None):
        """
        A decorator that apply marshalling to the return values of your methods.

        :param schema: The schema to be used to serialize the values.
        :param envelope: The key used to envelope the data.
        :return: A function.
        """
        schema = schema() if isclass(schema) else schema

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                data = func(*args, **kwargs)
                if isinstance(data, self.__app.response_class):
                    return data
                return marshal(data, schema, envelope)
            return wrapper
        return decorator

    def make_response(self, data):
        """
        Creates a Flask response object from the specified data.
        The appropriated encoder is taken based on the request header Accept.
        If there is not data to be serialized the response status code is 204.

        :param data: The Python object to be serialized.
        :return: A Flask response object.
        """

        status = headers = None
        if isinstance(data, tuple):
            data, status, headers = unpack(data)

        if data is None:
            data = self.__app.response_class(status=204)
        elif not isinstance(data, self.__app.response_class):
            media_type = request.accept_mimetypes.best_match(self.encoders, default=self.default_encoder)
            encoder = self.encoders.get(media_type)

            if encoder is None:
                raise InternalServerError()

            data_bytes = encoder(data)

            data = self.__app.response_class(data_bytes, mimetype=media_type)

        if status is not None:
            data.status_code = status

        if headers:
            data.headers.extend(headers)

        return data

    def trace_inspect(self):
        """
        A decorator that allows to inspect/change the trace data.
        """
        def decorator(f):
            self.tracer.inspector = f
            return f
        return decorator

    def trace_emit(self):
        """
        A decorator that allows to change the trace emitter.
        By default Python logging is used to emit the trace data.
        """
        def decorator(f):
            self.tracer.emitter = f
            return f
        return decorator

    def __decode_data(self, data):
        mimetype = get_best_match_for_content_type(self.decoders, self.default_decoder)

        if not mimetype:
            raise UnsupportedMediaType('Content-Type is not supported: ' + request.headers['content-type'])

        decoder = self.decoders.get(mimetype)

        try:
            return decoder(data)
        except:
            raise BadRequest('Invalid payload format.')

    def __from_source(self, param_name, field, getter_data, location):
        field = field() if isclass(field) else field
        if not field.required:
            field.allow_none = True

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                kwargs[param_name] = self.__parse_field(param_name, field, getter_data(), location)
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def __handle_error(self, e):
        if isinstance(e, HTTPException):
            code = e.code
            error = getattr(e, 'description', http_status_message(code))
        elif isinstance(e, ValidationError):
            code = 400
            error = validation_error_to_errors(e)
        else:
            code = 500
            error = str(e) if self.__app.config.get('DEBUG') else http_status_message(code)
            self.logger.error(str(e))

        errors = errors_to_dict(error)

        return self.make_response((errors, code))

    def __parse_field(self, field_name, field, data, location):
        field.allow_none = True

        field_name = field.load_from or field_name

        if isinstance(field, fields.List):
            raw_value = data.getlist(field_name) or missing
        else:
            raw_value = data.get(field_name) or missing

        if raw_value is missing:
            missing_value = field.missing
            raw_value = missing_value() if callable(missing_value) else missing_value

        if raw_value is missing and not field.required:
            raw_value = None

        try:
            return field.deserialize(raw_value, field_name, data)
        except ValidationError as e:
            e.messages = {field_name: e.messages}
            e.kwargs['location'] = location
            raise

    def __parse_body(self, schema):
        if not request.data:
            raise BadRequest('Payload is missing.')

        try:
            decoded_data = self.__decode_data(request.data)
        except:
            raise BadRequest('Invalid payload format.')

        model, errors = schema.load(decoded_data)

        if errors:
            raise ValidationError(errors, data=request.data, location='body')

        return model

    def __process_request(self, func, should_trace):
        def decorator(**kwargs):
            latency = response = error = None

            if should_trace and self.tracer.enabled:
                latency = Stopwatch.start_new()

            try:
                response = func(**kwargs)
                response = self.make_response(response)
                return response
            except Exception as e:
                error = e
                response = self.__handle_error(e)
                return response
            finally:
                if should_trace and self.tracer.enabled:
                    latency.stop()
                    self.tracer.trace(request, response, error, latency)

        return decorator

    def __setup(self):
        for endpoint in self.__app.view_functions.keys():
            should_trace = False

            for rule in self.__app.url_map.iter_rules(endpoint):
                if self.tracer.match(rule):
                    should_trace = True
                    break

            self.__app.view_functions[endpoint] = \
                self.__process_request(self.__app.view_functions[endpoint], should_trace)



"""Encoders and decoders used to deserialize the request body and serialize the response body."""

try:
    import simplejson as json
except ImportError:
    import json


def json_decode(data):
    """
    Deserializes a byte array containing a JSON document to a Python object.
    :param data: A byte array containing a JSON document.
    :return: A Python object.
    """
    return json.loads(data.decode())


def json_encode(data):
    """
    Serializes a Python object into a byte array containing a JSON document.
    :param data: A Python object.
    :return: A byte array containing a JSON document.
    """
    return json.dumps(data).encode()



from setuptools import setup

setup(
    name='flask-io',
    version='1.7.0',
    packages=['flask_io'],
    url='https://github.com/viniciuschiele/flask-io',
    license='MIT',
    author='Vinicius Chiele',
    author_email='vinicius.chiele@gmail.com',
    description='Flask-IO is a library for parsing Flask request arguments into parameters and for serialization of complex objects into Flask response.',
    keywords=['flask', 'rest', 'parse', 'encode', 'decode', 'request', 'json', 'marshmallow'],
    install_requires=['flask==0.10.1', 'python-dateutil>=2.4.2', 'marshmallow==2.2.0'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython'
    ]
)
